package gopher

import (
	"gophi/core"
	"os"
	"strings"
)

// Gophermap line formatting constants
const (
	maxSelectorLen = 255
	nullHost       = "null.host"
	nullPort       = "0"
	errorSelector  = "/error_selector_length"
)

// formatName formats a gopher line name string
func formatName(name string) string {
	if len(name) > pageWidth {
		return name[:pageWidth-4] + "...\t"
	}
	return name + "\t"
}

// formatSelector formats a gopher line selector string
func formatSelector(selector string) string {
	if len(selector) > maxSelectorLen {
		return errorSelector
	}
	return selector
}

// replacePlacementStrs replaces any placement strings found in the line (e.g $hostname, $port)
func replacePlacementStrs(line string) string {
	split := strings.SplitN(line, "\t", 4)

	// Either invalid line, or information line without host + port
	if len(split) != 4 {
		return line
	}

	// Return recombined (and replaced! string)
	return split[0] + "\t" +
		split[1] + "\t" +
		strings.Replace(split[2], "$host", core.Hostname, 1) + "\t" +
		strings.Replace(split[3], "$port", core.Port, 1)
}

// buildLine builds a gopher line string
func buildLine(t ItemType, name, selector, host, port string) []byte {
	return []byte(string(t) + formatName(name) + formatSelector(selector) + "\t" + host + "\t" + port + "\r\n")
}

// buildInfoLine builds a gopher info line string
func buildInfoLine(line string) []byte {
	return []byte(string(typeInfo) + formatName(line) + "\t" + nullHost + "\t" + nullPort + "\r\n")
}

// buildErrorLine builds a gopher error line string
func buildErrorLine(selector string) []byte {
	return []byte(string(typeError) + selector + "\r\n" + ".\r\n")
}

// appendFileListing formats and appends a new file entry as part of a directory listing
func appendFileListing(b []byte, file os.FileInfo, p *core.Path) []byte {
	// Handle file type
	switch {
	case file.Mode().IsDir():
		return append(b, buildLine(typeDirectory, file.Name(), p.Selector()+"/", core.Hostname, core.Port)...)
	case file.Mode().IsRegular():
		t := getItemType(file.Name())
		return append(b, buildLine(t, file.Name(), p.Selector(), core.Hostname, core.Port)...)
	default:
		return b
	}
}

// buildFooter formats a raw gopher footer ready to attach to end of gophermaps (including DOS line-end)
func buildFooter(raw string) []byte {
	ret := make([]byte, 0)

	if raw != "" {
		ret = append(ret, buildInfoLine(footerLineSeparator())...)

		for _, line := range strings.Split(raw, "\n") {
			ret = append(ret, buildInfoLine(line)...)
		}
	}

	return append(ret, []byte(".\r\n")...)
}

// footerLineSeparator is an internal function that generates a footer line separator string
func footerLineSeparator() string {
	ret := ""
	for i := 0; i < pageWidth; i++ {
		ret += "_"
	}
	return ret
}
